#include "nv_dev.h"
#include "nv-freebsd.h"
#include "nv_ref.h"
#include <machine/bus.h>
#include <sys/bus.h>
#include <sys/proc.h>
#include <vm/vm_kern.h>
#include <vm/vm_extern.h>

static void nv_intr		__P((void *));
static int nv_open		__P((dev_t, int, int, struct proc *));
static void nv_interrupts_disable __P((nv_state_t *));
static void nv_interrupts_enable __P((nv_state_t *));
static int nv_close 		__P((dev_t, int, int, struct proc *));
static int nv_ioctl		__P((dev_t, unsigned long, caddr_t, int, struct proc *));
static int nv_real_ioctl	__P((dev_t, unsigned long, unsigned long *, int, struct proc *));
static int nv_mmap		__P((dev_t, vm_offset_t, int));
static int nv_poll		__P((dev_t, int, struct proc *));

/* XXX this should go in nv.h */
extern device_t	nv_devices[NV_MAX_DEVICES];

#define GET_STATE device_t device = nv_devices[minor(dev)]; \
  nv_state_t *sc_old = device ? nv_device_get_softc (device) : NULL; \
  nv_freebsd_state_t *sc = sc_old ? (nv_freebsd_state_t *)(sc_old->os_state) : NULL; \
  if (sc == NULL) return (ENXIO);
  

static void nv_intr (arg)
	void			*arg;
{
	nv_state_t		*nv;
	U032			need_to_run_bottom_half = 0;

	nv = (nv_state_t *)arg;
	RmIsr (nv->device_number, &need_to_run_bottom_half);

	if (need_to_run_bottom_half) {
		RmIsrBottomHalf (nv->pdev);
	}

}


/* Done, save the private file information.
 */
static int nv_open (dev, oflags, devtype, p)
	dev_t			dev;
	int			oflags;
	int			devtype;
	struct proc		*p;
{
	int			i;
	GET_STATE

	printf ("NV%d: open by process %d\n", minor(dev), p->p_pid);

	if (!(sc_old->flags & NV_FLAG_OPEN)) {
		if (nv_map_device (device, minor(dev)) < 0)
			goto fail;

		dvprintf ("NV%d: RmInitAdapter\n", minor(dev));
		if (!RmInitAdapter (sc_old)) {
			printf ("nvidia%d: RmInitAdapter failed\n", minor(dev));
			goto fail;
		}

		sc_old->flags |= NV_FLAG_POSTED;

		sc->intr_cookie = 0;

		bus_setup_intr (device, sc->intr.res, INTR_TYPE_MISC, nv_intr, sc_old, &sc->intr_cookie);

		sc_old->flags |= NV_FLAG_OPEN;

		for (i = 0; i < MAX_KEYS; ++i) {
			sc->allocs[i] = 0;
		}
	}

	sc_old->usage_count++;

	return (0);
fail:
	nv_release_resources (device);
	return (ENXIO);
}

static void nv_interrupts_disable (nv)
	nv_state_t		*nv;
{
	nv->regs[NV_PMC_INTR_EN_0 >> 2] = 0;
}

static void nv_interrupts_enable (nv)
	nv_state_t		*nv;
{
	nv->regs[NV_PMC_INTR_EN_0 >> 2] = 1;
}

/* done, save some locks
 */
static int nv_close (dev, oflags, devtype, p)
	dev_t			dev;
	int			oflags;
	int			devtype;
	struct proc		*p;
{
	struct nv_alloc		*ptr;
	int			i;
	GET_STATE

	dprintf ("NV%d: close\n", minor(dev));

	nv_unix_free_all_unused_clients (sc_old, p->p_pid, NULL); 

	if (--sc_old->usage_count != 0)
		return (0);

	nv_interrupts_disable (sc_old);

	if (sc->intr.res != NULL) {
		if (sc->intr_cookie != NULL)
			bus_teardown_intr (device, sc->intr.res, sc->intr_cookie);
		bus_release_resource(device, SYS_RES_IRQ, 0, sc->intr.res);
		sc->intr.res = NULL;
	}

	RmShutdownAdapter (sc_old);
	nv_release_resources (device);
	sc_old->flags &= ~(NV_FLAG_OPEN | NV_FLAG_WAITING);

	for (i = 0; i < MAX_KEYS; ++i) {
		ptr = sc->allocs[i];
		if (!ptr)
			continue;
		printf ("kmem_free: %08lx\n", ptr->vaddr);
		kmem_free (kernel_map, ptr->vaddr, ptoa(ptr->pages));
		free (ptr->page_table, M_NV);
		free (ptr, M_NV);
		sc->allocs[i] = 0;
	}

	return (0);
}

nv_freebsd_state_t		*lalala;

static int nv_real_ioctl (dev, cmd, addy, flag, p)
	dev_t			dev;
	unsigned long		cmd;
	unsigned long		*addy;
	int			flag;
	struct proc		*p;
{
	nv_ioctl_memory_vtop_t	params;
	int			state;
	int			status = EINVAL;
	GET_STATE

	if (inc_entry())
		return (EINVAL);

	lalala = sc;

	switch (cmd) {
	case NV_IOCTL_MODULE_RESET:
		dvprintf ("NV_IOCTL_MODULE_RESET\n");
		break;

	case NV_IOCTL_INTERRUPTS:
		dvprintf ("NV_IOCTL_INTERRUPTS\n");
		break;
		state = 0;
		if (state)
			nv_interrupts_enable (sc_old);
		else
			nv_interrupts_disable (sc_old);
		status = 0;
		break;

	case NV_IOCTL_MEMORY_VTOP:
		dvprintf ("NV_IOCTL_MEMORY_VTOP\n");
		if (copyin (addy, &params, sizeof(params))) {
			printf ("copyin failure\n");
			return (EINVAL);
		}
		status = nv_vtop(&params, 1);
		break;

	default:
		dprintf ("nvExtEscape\n");
		status = nvExtEscape(NULL, sc_old, cmd, addy) ? 0 : EINVAL;
		break;
	}

	return (status);
}

static int nv_ioctl (dev, cmd, data, flag, p)
	dev_t			dev;
	unsigned long		cmd;
	caddr_t			data;
	int			flag;
	struct proc		*p;
{
	unsigned long *		addy;

	if (cmd != DUMMY_IOCTL)
		return (EINVAL);

	addy = *(unsigned long **)data;

	return (nv_real_ioctl (dev, last_ioctl_command, addy, flag, p));
}

extern struct nv_alloc		*last;

static int nv_mmap (dev, offset, prot)
	dev_t			dev;
	vm_offset_t		offset;
	int			prot;
{
	int			aoffset;
	unsigned long		res = -1;
	GET_STATE

/*	if ((offset == 0x800000) && !inc_in_progress())
		inc_init (59);
*/
	last = NULL;

/*	printf ("nv_mmap: %08lx: ", offset); */
	if (offset >= NV_MMAP_REG_OFFSET && offset < NV_MMAP_FB_OFFSET) {
		aoffset = offset - NV_MMAP_REG_OFFSET;
		if (aoffset == 0)
			printf ("reg_physical\n");
		res = atop(sc_old->reg_physical_address + aoffset);
	} else if (offset >= NV_MMAP_FB_OFFSET && offset < NV_MMAP_ALLOCATION_OFFSET) {
		aoffset = offset - NV_MMAP_FB_OFFSET;
		if (aoffset == 0)
			printf ("fb_physical\n");
		res = atop(sc_old->fb_physical_address + aoffset);
	} else if (offset >= NV_MMAP_ALLOCATION_OFFSET && offset < NV_MMAP_AGP_OFFSET) {
		struct nv_alloc *ptr;
		unsigned char	key;

		key = NV_KEY_FROM_ALLOC (offset);
		dprintf ("alloc key: %d\n", key);
		if (key >= MAX_KEYS || key < 0)
			return (-1);

		aoffset = NV_ALLOC_OFFSET(offset);

		ptr = sc->allocs[key];
		if (!ptr) {
			printf ("couldn't find relevant nv_alloc\n");
			return (-1);
		}	

		if (aoffset == 0) {
			printf ("%08x\n", ptoa(ptr->pages));
		}

		if (aoffset >= ptoa(ptr->pages)) {
			printf ("mmap out of range\n");
			return (-1);
		}
		res = vtophys (ptr->vaddr + aoffset);
		res = atop (res);
		dvprintf ("%08lx\n", res);
		last = ptr;
	} else if (offset == NV_MMAP_AGP_OFFSET) {
		printf ("agp\n");
	}

	return (res);
}

static int nv_poll (dev, events, p)
	dev_t			dev;
	int			events;
	struct proc		*p;
{
	printf ("nv_poll\n");

	return (EINVAL);
}

struct cdevsw nvidia_cdevsw = {
	nv_open,
	nv_close,
	noread,
	nowrite,
	nv_ioctl,
	nv_poll,
	nv_mmap,
	nostrategy,
	"nvidia",
	NV_MAJOR_DEVICE_NUMBER,
	nodump,
	nopsize,
	D_MEM | D_TRACKCLOSE,
	-1
};
